@@ -3,29 +3,32 @@
 import copy
 import logging
 from decimal import Decimal
+from importlib import import_module
 
-from sql import Null, Column
+import stdnum
+import stdnum.exceptions
+from sql import Null, Column, Literal
+from sql.operators import Equal
 
+from trytond.i18n import gettext
 from trytond.model import (
-    ModelView, ModelSQL, Model, UnionMixin, DeactivableMixin, fields)
-from trytond.pyson import Eval
+    ModelView, ModelSQL, Model, UnionMixin, DeactivableMixin, sequence_ordered,
+    Exclude, fields)
+from trytond.pyson import Eval, If
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 from trytond import backend
-from trytond.config import config
+from trytond.tools import lstrip_wildcard
 from trytond.tools.multivalue import migrate_property
 from trytond.modules.company.model import (
     CompanyMultiValueMixin, CompanyValueMixin)
+from .exceptions import InvalidIdentifierCode
+from .ir import price_decimal
 
-__all__ = ['Template', 'Product', 'price_digits', 'TemplateFunction',
-    'ProductListPrice', 'ProductCostPriceMethod', 'ProductCostPrice',
-    'TemplateCategory', 'TemplateCategoryAll']
+
+__all__ = ['price_digits', 'round_price', 'TemplateFunction']
 logger = logging.getLogger(__name__)
 
-STATES = {
-    'readonly': ~Eval('active', True),
-    }
-DEPENDS = ['active']
 TYPES = [
     ('goods', 'Goods'),
     ('assets', 'Assets'),
@@ -36,48 +39,80 @@ COST_PRICE_METHODS = [
     ('average', 'Average'),
     ]
 
-price_digits = (16, config.getint('product', 'price_decimal', default=4))
+price_digits = (16, price_decimal)
+
+
+def round_price(value, rounding=None):
+    "Round price using the price digits"
+    if isinstance(value, int):
+        return Decimal(value)
+    return value.quantize(
+        Decimal(1) / 10 ** price_digits[1], rounding=rounding)
 
 
 class Template(
         DeactivableMixin, ModelSQL, ModelView, CompanyMultiValueMixin):
     "Product Template"
     __name__ = "product.template"
-    name = fields.Char('Name', size=None, required=True, translate=True,
-        select=True, states=STATES, depends=DEPENDS)
-    type = fields.Selection(TYPES, 'Type', required=True, states=STATES,
-        depends=DEPENDS)
+    _order_name = 'rec_name'
+    name = fields.Char(
+        "Name", size=None, required=True, translate=True, select=True)
+    code_readonly = fields.Function(
+        fields.Boolean("Code Readonly"), 'get_code_readonly')
+    code = fields.Char(
+        "Code", select=True,
+        states={
+            'readonly': Eval('code_readonly', False),
+            },
+        depends=['code_readonly'])
+    type = fields.Selection(TYPES, "Type", required=True)
     consumable = fields.Boolean('Consumable',
         states={
-            'readonly': ~Eval('active', True),
             'invisible': Eval('type', 'goods') != 'goods',
             },
-        depends=['active', 'type'])
+        depends=['type'],
+        help="Check to allow stock moves to be assigned "
+        "regardless of stock level.")
     list_price = fields.MultiValue(fields.Numeric(
             "List Price", required=True, digits=price_digits,
-            states=STATES, depends=DEPENDS))
+            help="The standard price the product is sold at."))
     list_prices = fields.One2Many(
         'product.list_price', 'template', "List Prices")
     cost_price = fields.Function(fields.Numeric(
-            "Cost Price", digits=price_digits), 'get_cost_price')
+            "Cost Price", digits=price_digits,
+            help="The amount it costs to purchase or make the product, "
+            "or carry out the service."),
+        'get_cost_price')
     cost_price_method = fields.MultiValue(fields.Selection(
             COST_PRICE_METHODS, "Cost Price Method", required=True,
-            states=STATES, depends=DEPENDS))
+            help="The method used to calculate the cost price."))
     cost_price_methods = fields.One2Many(
         'product.cost_price_method', 'template', "Cost Price Methods")
-    default_uom = fields.Many2One('product.uom', 'Default UOM', required=True,
-        states=STATES, depends=DEPENDS)
+    default_uom = fields.Many2One('product.uom', "Default UOM", required=True,
+        help="The standard unit of measure for the product.\n"
+        "Used internally when calculating the stock levels of goods "
+        "and assets.")
     default_uom_category = fields.Function(
         fields.Many2One('product.uom.category', 'Default UOM Category'),
         'on_change_with_default_uom_category',
         searcher='search_default_uom_category')
-    categories = fields.Many2Many('product.template-product.category',
-        'template', 'category', 'Categories', states=STATES, depends=DEPENDS)
+    default_uom_digits = fields.Function(fields.Integer("Default Unit Digits"),
+        'on_change_with_default_uom_digits')
+    categories = fields.Many2Many(
+        'product.template-product.category', 'template', 'category',
+        "Categories",
+        help="The categories that the product is in.\n"
+        "Used to group similar products together.")
     categories_all = fields.Many2Many(
         'product.template-product.category.all',
         'template', 'category', "Categories", readonly=True)
-    products = fields.One2Many('product.product', 'template', 'Variants',
-        states=STATES, depends=DEPENDS)
+    products = fields.One2Many(
+        'product.product', 'template', "Variants",
+        domain=[
+            If(~Eval('active'), ('active', '=', False), ()),
+            ],
+        depends=['active'],
+        help="The different variants the product comes in.")
 
     @classmethod
     def __register__(cls, module_name):
@@ -91,6 +126,11 @@ class Template(
                 'The column "category" on table "%s" must be dropped manually',
                 cls._table)
 
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls._order.insert(0, ('rec_name', 'ASC'))
+
     @classmethod
     def multivalue_model(cls, field):
         pool = Pool()
@@ -100,6 +140,34 @@ class Template(
             return pool.get('product.cost_price_method')
         return super(Template, cls).multivalue_model(field)
 
+    @classmethod
+    def order_rec_name(cls, tables):
+        table, _ = tables[None]
+        return [table.code, table.name]
+
+    def get_rec_name(self, name):
+        if self.code:
+            return '[' + self.code + '] ' + self.name
+        else:
+            return self.name
+
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        if clause[1].startswith('!') or clause[1].startswith('not '):
+            bool_op = 'AND'
+        else:
+            bool_op = 'OR'
+        code_value = clause[2]
+        if clause[1].endswith('like'):
+            code_value = lstrip_wildcard(clause[2])
+        return [bool_op,
+            ('name',) + tuple(clause[1:]),
+            ('code', clause[1], code_value) + tuple(clause[3:]),
+            ('products.code', clause[1], code_value) + tuple(clause[3:]),
+            ('products.identifiers.code', clause[1], code_value)
+            + tuple(clause[3:]),
+            ]
+
     @staticmethod
     def default_type():
         return 'goods'
@@ -120,12 +188,24 @@ class Template(
         return Configuration(1).get_multivalue(
             'default_cost_price_method', **pattern)
 
-    @staticmethod
-    def default_products():
-        if Transaction().user == 0:
+    @classmethod
+    def default_products(cls):
+        transaction = Transaction()
+        if (transaction.user == 0
+                or not transaction.context.get('default_products', True)):
             return []
         return [{}]
 
+    @classmethod
+    def default_code_readonly(cls):
+        pool = Pool()
+        Configuration = pool.get('product.configuration')
+        config = Configuration(1)
+        return bool(config.template_sequence)
+
+    def get_code_readonly(self, name):
+        return self.default_code_readonly()
+
     @fields.depends('type', 'cost_price_method')
     def on_change_type(self):
         if self.type == 'service':
@@ -141,12 +221,51 @@ class Template(
         return [('default_uom.category' + clause[0].lstrip(name),)
             + tuple(clause[1:])]
 
+    @fields.depends('default_uom')
+    def on_change_with_default_uom_digits(self, name=None):
+        if self.default_uom:
+            return self.default_uom.digits
+
+    @classmethod
+    def _new_code(cls):
+        pool = Pool()
+        Configuration = pool.get('product.configuration')
+        config = Configuration(1)
+        sequence = config.template_sequence
+        if sequence:
+            return sequence.get()
+
     @classmethod
     def create(cls, vlist):
+        pool = Pool()
+        Product = pool.get('product.product')
         vlist = [v.copy() for v in vlist]
         for values in vlist:
             values.setdefault('products', None)
-        return super(Template, cls).create(vlist)
+            if not values.get('code'):
+                values['code'] = cls._new_code()
+        templates = super(Template, cls).create(vlist)
+        products = sum((t.products for t in templates), ())
+        Product.sync_code(products)
+        return templates
+
+    @classmethod
+    def write(cls, *args):
+        pool = Pool()
+        Product = pool.get('product.product')
+        super().write(*args)
+        templates = sum(args[0:None:2], [])
+        products = sum((t.products for t in templates), ())
+        Product.sync_code(products)
+
+    @classmethod
+    def copy(cls, templates, default=None):
+        if default is None:
+            default = {}
+        else:
+            default = default.copy()
+        default.setdefault('code', None)
+        return super().copy(templates, default=default)
 
     @classmethod
     def search_global(cls, text):
@@ -182,30 +301,61 @@ class TemplateFunction(fields.Function):
                 tables['template'] = {
                     None: (template, product.template == template.id),
                     }
-            else:
-                template = tables['template']
             return getattr(Template, name).convert_order(
                 name, tables['template'], Template)
         return order
 
+    def definition(self, model, language):
+        pool = Pool()
+        Template = pool.get('product.template')
+        definition = super().definition(model, language)
+        definition['searchable'] = self._field.definition(
+            Template, language)['searchable']
+        return definition
+
 
 class Product(
         DeactivableMixin, ModelSQL, ModelView, CompanyMultiValueMixin):
     "Product Variant"
     __name__ = "product.product"
     _order_name = 'rec_name'
-    template = fields.Many2One('product.template', 'Product Template',
-        required=True, ondelete='CASCADE', select=True, states=STATES,
-        depends=DEPENDS)
-    code = fields.Char("Code", size=None, select=True, states=STATES,
-        depends=DEPENDS)
+    template = fields.Many2One(
+        'product.template', "Product Template",
+        required=True, ondelete='CASCADE', select=True,
+        search_context={'default_products': False},
+        domain=[
+            If(Eval('active'), ('active', '=', True), ()),
+            ],
+        depends=['active'],
+        help="The product that defines the common properties "
+        "inherited by the variant.")
+    code_readonly = fields.Function(fields.Boolean('Code Readonly'),
+        'get_code_readonly')
+    prefix_code = fields.Function(fields.Char(
+            "Prefix Code",
+            states={
+                'invisible': ~Eval('prefix_code'),
+                }),
+        'on_change_with_prefix_code')
+    suffix_code = fields.Char(
+        "Suffix Code",
+        states={
+            'readonly': Eval('code_readonly', False),
+            },
+        depends=['code_readonly'],
+        help="The unique identifier for the product (aka SKU).")
+    code = fields.Char("Code", readonly=True, select=True,
+        help="A unique identifier for the variant.")
+    identifiers = fields.One2Many(
+        'product.identifier', 'product', "Identifiers",
+        help="Other identifiers associated with the variant.")
     cost_price = fields.MultiValue(fields.Numeric(
             "Cost Price", required=True, digits=price_digits,
-            states=STATES, depends=DEPENDS))
+            help="The amount it costs to purchase or make the variant, "
+            "or carry out the service."))
     cost_prices = fields.One2Many(
         'product.cost_price', 'product', "Cost Prices")
-    description = fields.Text("Description", translate=True, states=STATES,
-        depends=DEPENDS)
+    description = fields.Text("Description", translate=True)
     list_price_uom = fields.Function(fields.Numeric('List Price',
         digits=price_digits), 'get_price_uom')
     cost_price_uom = fields.Function(fields.Numeric('Cost Price',
@@ -221,6 +371,16 @@ class Product(
         cls._no_template_field.update(['products'])
 
         super(Product, cls).__setup__()
+        cls.__access__.add('template')
+        cls._order.insert(0, ('rec_name', 'ASC'))
+
+        t = cls.__table__()
+        cls._sql_constraints = [
+            ('code_exclude', Exclude(t, (t.code, Equal),
+                    where=(t.active == Literal(True))
+                    & (t.code != '')),
+                'product.msg_product_code_unique'),
+            ]
 
         for attr in dir(Template):
             tfield = getattr(Template, attr)
@@ -230,7 +390,17 @@ class Product(
                 continue
             field = getattr(cls, attr, None)
             if not field or isinstance(field, TemplateFunction):
-                setattr(cls, attr, TemplateFunction(copy.deepcopy(tfield)))
+                tfield = copy.deepcopy(tfield)
+                if hasattr(tfield, 'field'):
+                    tfield.field = None
+                invisible_state = ~Eval('template')
+                if 'invisible' in tfield.states:
+                    tfield.states['invisible'] |= invisible_state
+                else:
+                    tfield.states['invisible'] = invisible_state
+                if 'template' not in tfield.depends:
+                    tfield.depends.append('template')
+                setattr(cls, attr, TemplateFunction(tfield))
                 order_method = getattr(cls, 'order_%s' % attr, None)
                 if (not order_method
                         and not isinstance(tfield, (
@@ -239,6 +409,29 @@ class Product(
                                 fields.Many2Many))):
                     order_method = TemplateFunction.order(attr)
                     setattr(cls, 'order_%s' % attr, order_method)
+                if isinstance(tfield, fields.One2Many):
+                    getattr(cls, attr).setter = '_set_template_function'
+
+    @classmethod
+    def __register__(cls, module):
+        table = cls.__table__()
+        table_h = cls.__table_handler__(module)
+        fill_suffix_code = (
+            table_h.column_exist('code')
+            and not table_h.column_exist('suffix_code'))
+        super().__register__(module)
+        cursor = Transaction().connection.cursor()
+
+        # Migration from 5.4: split code into prefix/suffix
+        if fill_suffix_code:
+            cursor.execute(*table.update(
+                    [table.suffix_code],
+                    [table.code]))
+
+    @classmethod
+    def _set_template_function(cls, products, name, value):
+        # Prevent NotImplementedError for One2Many
+        pass
 
     @fields.depends('template', '_parent_template.id')
     def on_change_template(self):
@@ -253,6 +446,9 @@ class Product(
     def get_template(self, name):
         value = getattr(self.template, name)
         if isinstance(value, Model):
+            field = getattr(self.__class__, name)
+            if field._type == 'reference':
+                return str(value)
             return value.id
         elif (isinstance(value, (list, tuple))
                 and value and isinstance(value[0], Model)):
@@ -260,6 +456,11 @@ class Product(
         else:
             return value
 
+    @fields.depends('template', '_parent_template.code')
+    def on_change_with_prefix_code(self, name=None):
+        if self.template:
+            return self.template.code
+
     @classmethod
     def multivalue_model(cls, field):
         pool = Pool()
@@ -302,9 +503,14 @@ class Product(
             bool_op = 'AND'
         else:
             bool_op = 'OR'
+        code_value = clause[2]
+        if clause[1].endswith('like'):
+            code_value = lstrip_wildcard(clause[2])
         return [bool_op,
-            ('code',) + tuple(clause[1:]),
+            ('code', clause[1], code_value) + tuple(clause[3:]),
+            ('identifiers.code', clause[1], code_value) + tuple(clause[3:]),
             ('template.name',) + tuple(clause[1:]),
+            ('template.code', clause[1], code_value) + tuple(clause[3:]),
             ]
 
     @staticmethod
@@ -331,18 +537,84 @@ class Product(
             icon = icon or 'tryton-product'
             yield id_, rec_name, icon
 
+    @classmethod
+    def default_code_readonly(cls):
+        pool = Pool()
+        Configuration = pool.get('product.configuration')
+        config = Configuration(1)
+        return bool(config.product_sequence)
+
+    def get_code_readonly(self, name):
+        return self.default_code_readonly()
+
+    @classmethod
+    def _new_suffix_code(cls):
+        pool = Pool()
+        Configuration = pool.get('product.configuration')
+        config = Configuration(1)
+        sequence = config.product_sequence
+        if sequence:
+            return sequence.get()
+
+    @classmethod
+    def create(cls, vlist):
+        vlist = [x.copy() for x in vlist]
+        for values in vlist:
+            if not values.get('suffix_code'):
+                values['suffix_code'] = cls._new_suffix_code()
+        products = super().create(vlist)
+        cls.sync_code(products)
+        return products
+
+    @classmethod
+    def write(cls, *args):
+        super().write(*args)
+        products = sum(args[0:None:2], [])
+        cls.sync_code(products)
+
+    @classmethod
+    def copy(cls, products, default=None):
+        if default is None:
+            default = {}
+        else:
+            default = default.copy()
+        default.setdefault('suffix_code', None)
+        default.setdefault('code', None)
+        return super().copy(products, default=default)
+
+    @property
+    def list_price_used(self):
+        transaction = Transaction()
+        with transaction.reset_context(), \
+                transaction.set_context(self._context):
+            return self.template.get_multivalue('list_price')
+
+    @classmethod
+    def sync_code(cls, products):
+        for product in products:
+            code = ''.join(filter(None, [
+                        product.prefix_code, product.suffix_code]))
+            if not code:
+                code = None
+            if code != product.code:
+                product.code = code
+        cls.save(products)
+
 
 class ProductListPrice(ModelSQL, CompanyValueMixin):
     "Product List Price"
     __name__ = 'product.list_price'
     template = fields.Many2One(
-        'product.template', "Template", ondelete='CASCADE', select=True)
+        'product.template', "Template", ondelete='CASCADE', select=True,
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
     list_price = fields.Numeric("List Price", digits=price_digits)
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        exist = TableHandler.table_exist(cls._table)
+        exist = backend.TableHandler.table_exist(cls._table)
 
         super(ProductListPrice, cls).__register__(module_name)
 
@@ -363,7 +635,11 @@ class ProductCostPriceMethod(ModelSQL, CompanyValueMixin):
     "Product Cost Price Method"
     __name__ = 'product.cost_price_method'
     template = fields.Many2One(
-        'product.template', "Template", ondelete='CASCADE', select=True)
+        'product.template', "Template", ondelete='CASCADE', select=True,
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
     cost_price_method = fields.Selection(
         'get_cost_price_methods', "Cost Price Method")
 
@@ -371,13 +647,13 @@ class ProductCostPriceMethod(ModelSQL, CompanyValueMixin):
     def __register__(cls, module_name):
         pool = Pool()
         ProductCostPrice = pool.get('product.cost_price')
-        TableHandler = backend.get('TableHandler')
         sql_table = cls.__table__()
         cost_price = ProductCostPrice.__table__()
         cursor = Transaction().connection.cursor()
 
-        exist = TableHandler.table_exist(cls._table)
-        cost_price_exist = TableHandler.table_exist(ProductCostPrice._table)
+        exist = backend.TableHandler.table_exist(cls._table)
+        cost_price_exist = backend.TableHandler.table_exist(
+            ProductCostPrice._table)
 
         super(ProductCostPriceMethod, cls).__register__(module_name)
 
@@ -385,7 +661,8 @@ class ProductCostPriceMethod(ModelSQL, CompanyValueMixin):
         if not exist and not cost_price_exist:
             cls._migrate_property([], [], [])
         elif not exist and cost_price_exist:
-            cost_price_table = TableHandler(ProductCostPrice, module_name)
+            cost_price_table = backend.TableHandler(
+                ProductCostPrice, module_name)
             if cost_price_table.column_exist('template'):
                 columns = ['create_uid', 'create_date',
                     'write_uid', 'write_date',
@@ -418,7 +695,11 @@ class ProductCostPrice(ModelSQL, CompanyValueMixin):
     "Product Cost Price"
     __name__ = 'product.cost_price'
     product = fields.Many2One(
-        'product.product', "Product", ondelete='CASCADE', select=True)
+        'product.product', "Product", ondelete='CASCADE', select=True,
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
     cost_price = fields.Numeric(
         "Cost Price", digits=price_digits)
 
@@ -426,12 +707,11 @@ class ProductCostPrice(ModelSQL, CompanyValueMixin):
     def __register__(cls, module_name):
         pool = Pool()
         Product = pool.get('product.product')
-        TableHandler = backend.get('TableHandler')
         sql_table = cls.__table__()
         product = Product.__table__()
         cursor = Transaction().connection.cursor()
 
-        exist = TableHandler.table_exist(cls._table)
+        exist = backend.TableHandler.table_exist(cls._table)
 
         super(ProductCostPrice, cls).__register__(module_name)
 
@@ -488,3 +768,62 @@ class TemplateCategoryAll(UnionMixin, ModelSQL):
     @classmethod
     def union_models(cls):
         return ['product.template-product.category']
+
+
+class ProductIdentifier(sequence_ordered(), ModelSQL, ModelView):
+    "Product Identifier"
+    __name__ = 'product.identifier'
+    _rec_name = 'code'
+    product = fields.Many2One('product.product', "Product", ondelete='CASCADE',
+        required=True, select=True,
+        help="The product identified by the code.")
+    type = fields.Selection([
+            (None, ''),
+            ('ean', "International Article Number"),
+            ('isan', "International Standard Audiovisual Number"),
+            ('isbn', "International Standard Book Number"),
+            ('isil', "International Standard Identifier for Libraries"),
+            ('isin', "International Securities Identification Number"),
+            ('ismn', "International Standard Music Number"),
+            ], "Type")
+    type_string = type.translated('type')
+    code = fields.Char("Code", required=True)
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.add('product')
+
+    @fields.depends('type', 'code')
+    def on_change_with_code(self):
+        if self.type and self.type != 'other':
+            try:
+                module = import_module('stdnum.%s' % self.type)
+                return module.compact(self.code)
+            except ImportError:
+                pass
+            except stdnum.exceptions.ValidationError:
+                pass
+        return self.code
+
+    def pre_validate(self):
+        super().pre_validate()
+        self.check_code()
+
+    @fields.depends('type', 'product', 'code')
+    def check_code(self):
+        if self.type:
+            try:
+                module = import_module('stdnum.%s' % self.type)
+            except ModuleNotFoundError:
+                return
+            if not module.is_valid(self.code):
+                if self.product and self.product.id > 0:
+                    product = self.product.rec_name
+                else:
+                    product = ''
+                raise InvalidIdentifierCode(
+                    gettext('product.msg_invalid_code',
+                        type=self.type_string,
+                        code=self.code,
+                        product=product))
